#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['orchestrator',
 'orchestrator.api',
 'orchestrator.api.api_v1',
 'orchestrator.api.api_v1.endpoints',
 'orchestrator.cli',
 'orchestrator.config',
 'orchestrator.db',
 'orchestrator.domain',
 'orchestrator.forms',
 'orchestrator.schedules',
 'orchestrator.schemas',
 'orchestrator.services',
 'orchestrator.utils',
 'orchestrator.workflows',
 'orchestrator.workflows.tasks']

package_data = \
{'': ['*'],
 'orchestrator': ['migrations/*',
                  'migrations/templates/*',
                  'migrations/versions/*',
                  'migrations/versions/schema/*'],
 'orchestrator.workflows': ['translations/*']}

install_requires = \
['aiocache==0.11.1',
 'aioredis==1.3.1',
 'alembic~=1.5.4',
 'click~=7.1',
 'colorama~=0.4.4',
 'deepmerge~=0.1.0',
 'fastapi~=0.63.0',
 'fastapi-etag~=0.2.2',
 'gunicorn~=20.0.4',
 'httptools==0.1.1',
 'more-itertools~=8.7.0',
 'itsdangerous~=1.1.0',
 'opentelemetry-api~=0.17b0',
 'opentelemetry-exporter-jaeger~=0.17b0',
 'opentelemetry-instrumentation-fastapi~=0.17b0',
 'opentelemetry-instrumentation-psycopg2~=0.17b0',
 'opentelemetry-instrumentation-redis~=0.17b0',
 'opentelemetry-instrumentation-requests~=0.17b0',
 'opentelemetry-instrumentation-sqlalchemy~=0.17b0',
 'opentelemetry-sdk==0.17b0',
 'psycopg2-binary~=2.8.6',
 'pydantic[email]~=1.7.3',
 'pynso-restconf~=2.1.0',
 'python-rapidjson==1.0',
 'pytz',
 'regex~=2020.11.13',
 'requests~=2.25.1',
 'ruamel.yaml~=0.16.12',
 'schedule~=1.0.0',
 'sentry-sdk',
 'SQLAlchemy~=1.3.24',
 'SQLAlchemy-Searchable~=1.2.0',
 'structlog~=20.2.0',
 'typer-cli~=0.0.11',
 'uvicorn[standard]~=0.13.3',
 'nwa-stdlib~=1.2.0',
 'oauth2-lib~=1.0.4']

extras_require = \
{'dev': ['bumpversion',
         'mypy_extensions',
         'pre-commit',
         'pydocstyle==3.0.0',
         'python-dotenv',
         'watchdog'],
 'doc': ['pystache~=0.5.4',
         'recommonmark~=0.7.1',
         'Sphinx~=2.3.1',
         'sphinx-autobuild~=2020.9.1',
         'sphinx-rtd-theme==0.5.0',
         'sphinxcontrib-apidoc~=0.3.0',
         'sphinxcontrib-blockdiag~=2.0.0',
         'sphinxcontrib-seqdiag~=2.0.0'],
 'test': ['apache-license-check',
          'black',
          'blinker',
          'deepdiff',
          'flake8',
          'flake8-bandit',
          'flake8-bugbear',
          'flake8-comprehensions',
          'flake8-docstrings',
          'flake8-junit-report',
          'flake8-logging-format',
          'flake8-pep3101',
          'flake8-print',
          'flake8-rst',
          'flake8-rst-docstrings',
          'flake8-tidy-imports',
          'isort',
          'jsonref',
          'mypy',
          'pytest',
          'pytest-cov',
          'pytest-httpx==0.10.1',
          'pytest-xdist',
          'requests-mock',
          'urllib3_mock']}

setup(name='orchestrator-core',
      version='0.0.14',
      description='This is the orchestrator workflow engine.',
      author='SURF',
      author_email='automation-beheer@surf.nl',
      url='https://github.com/workfloworchestrator/orchestrator-core',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6,<3.9',
     )
